/*
 * Support Vector Registration (SVR): svgm
 * A class for the Support Vector-parametrised Gaussian Mixture (SVGM)
 * data representation. Includes methods for loading point-sets, constructing
 * the SVGM and manipulation.
 * 
 * Campbell, D., and Petersson, L., “An Adaptive Data Representations for
 * Robust Point-Set Registration and Merging”, International Conference on
 * Computer Vision (ICCV), Santiago, Chile, IEEE, Dec. 2015
 * 
 * For the full license, see license.txt in the root directory
 * 
 * Author: Dylan Campbell
 * Date: 20150820
 * Revision: 1
 */

#include "svgm.h"

void DisableSVMOutput(const char *s){}

SVGM::SVGM()
{
	num_points_x_ = -1; num_points_y_ = -1;
	num_components_x_ = -1; num_components_y_ = -1;
	max_num_points_ = 20000;
	dimension_ = -1;
	sigma_ = 0.01;
	gamma_ = 1000;
	nu_ = 0.01;
//	durations_;
//	problem_x_;
//	problem_y_;
//	svm_parameters_;
	model_x_ = NULL;
	model_y_ = NULL;
//	phi_x_, phi_y_;
//	mu_x_, mu_y_;
	InitialiseSVMParameters();
}

SVGM::~SVGM()
{
	delete[] problem_x_.y;
	for(int i = 0; i < problem_x_.l; ++i)
		delete[] (problem_x_.x + i)->values;
	delete[] problem_x_.x;

	delete[] problem_y_.y;
	for(int i = 0; i < problem_y_.l; ++i)
		delete[] (problem_y_.x + i)->values;
	delete[] problem_y_.x;

	svm_free_and_destroy_model(&model_x_);
	svm_free_and_destroy_model(&model_y_);
}

/*
 * Public Class Functions
 */

int SVGM::LoadPointSetX(std::string filename)
{
	num_points_x_ = LoadPointSet(filename, &problem_x_);
	dimension_ = problem_x_.x[0].dim;
	return num_points_x_;
}

int SVGM::LoadPointSetY(std::string filename)
{
	num_points_y_ = LoadPointSet(filename, &problem_y_);
	if(dimension_ != problem_y_.x[0].dim)
	{
		std::cout << "Data point-set has a different dimension to the model point-set" << std::endl;
		num_points_y_ = -1;
	}
	return num_points_y_;
}

void SVGM::BuildSVGMs()
{
	clock_t  clock_begin;

	// Check that point-sets have been loaded
	if(num_points_x_ <= 0 || num_points_y_ <= 0)
	{
		std::cout << "Load Point-Sets Before Calling Run()" << std::endl;
		return;
	}

	// Set SVM Parameters (gamma and nu)
	InitialiseSVM();

	// Train One-Class SVMs
	clock_begin = clock();
	if(!TrainSVM(&problem_x_, &model_x_)) return;
	durations_[0] = (double)(clock() - clock_begin) / CLOCKS_PER_SEC;
	clock_begin = clock();
	if(!TrainSVM(&problem_y_, &model_y_)) return;
	durations_[1] = (double)(clock() - clock_begin) / CLOCKS_PER_SEC;

	// Map SVMs to GMMs
	clock_begin = clock();
	MapToGMMs();
	durations_[2] = (double)(clock() - clock_begin) / CLOCKS_PER_SEC;
}

/*
 * Private Class Functions
 */

void SVGM::InitialiseSVMParameters()
{
	svm_parameters_.svm_type = ONE_CLASS;
	svm_parameters_.kernel_type = RBF;
	svm_parameters_.gamma = 3200;
	svm_parameters_.nu = 0.01;
	svm_parameters_.cache_size = 1000;
	svm_parameters_.eps = 1e-3;
	svm_parameters_.shrinking = 0; // 1: Use shrinking heuristics
	svm_parameters_.probability = 0; // 1: Provide probability estimates
	svm_parameters_.degree = 0; // Degree of polynomial kernel

	svm_set_print_string_function(&DisableSVMOutput); // Comment out if output is desired
}

/*
 * Load whitespace-separated point-set from a file
 * Determines point-set size and dimension automatically
 */
int SVGM::LoadPointSet(std::string filename, struct svm_problem* problem)
{
	int num_rows = -1;
	int num_cols = -1;
	int num_points = -1;
	int num_dimensions = -1;
	std::ifstream ifile;
	vnl_matrix<double> pointset;
	vnl_matrix<double> pointset_full;

	ifile.open(filename.c_str(), std::ios_base::in);

	if(ifile.is_open())
	{
		if(pointset_full.read_ascii(ifile))
		{
			num_rows = pointset_full.rows();
			num_cols = pointset_full.cols();

			if(max_num_points_ == 0 || num_rows <= max_num_points_)
			{
				num_points = num_rows;
				num_dimensions = num_cols;
				pointset = pointset_full;
			}
			else
			{
				// Generate randomised indices
				std::vector<int> random_indices;
				for(int i = 0; i < num_rows; ++i) random_indices.push_back(i);
				std::random_shuffle(random_indices.begin(), random_indices.end());
				num_points = std::min(num_rows, max_num_points_);
				num_dimensions = num_cols;

				// Fill in pointset
				pointset.set_size(num_points, num_dimensions);
				for(int i = 0; i < num_points; i++)
				{
					for(int j = 0; j < num_dimensions; ++j)
					{
						pointset[i][j] = pointset_full[random_indices[i]][j];
					}
				}
			}
			PrepareSVMProblem(pointset, num_points, num_dimensions, problem);
		}
		else
		{
			std::cout << "Cannot parse input file '" << filename << "'" << std::endl;
		}
		ifile.close();
	}
	else
	{
		std::cout << "Unable to open point file '" << filename << "'" << std::endl;
	}

	return num_points;
}

/*
 * Load whitespace-separated point-set from a file
 * Determines point-set size and dimension automatically
 * Less robust than vnl method
 */
/*
int SVGM::LoadPointSet(std::string filename, struct svm_problem* problem)
{
	int num_rows = -1;
	int num_cols = -1;
	std::string line, s;
	std::ifstream ifile;
	std::istringstream ss;

	ifile.open(filename.c_str(), std::ios_base::in);

	if(ifile.is_open())
	{
		// Find the number of rows
		num_rows = std::count(std::istreambuf_iterator<char>(ifile), std::istreambuf_iterator<char>(), '\n'); // Includes newlines at end of document

		// Find the number of columns
		ifile.seekg(0); // Return to beginning of file
		getline(ifile, line);
		ss.str(line);
		num_cols = 0;
		while(ss >> s)
		{
			num_cols++;
		}

		// Read Point-Set
		problem->l = num_rows;
		problem->y = new double[problem->l];
		problem->x = new struct svm_node[problem->l];

		ifile.seekg(0); // Return to beginning of file
		for(int i = 0; i < problem->l; i++)
		{
			problem->y[i] = 1; // Class Label
			(problem->x + i)->dim = num_cols; // Dimension
			(problem->x + i)->values = new double[num_cols];
			for(int j = 0; j < num_cols; ++j)
			{
				ifile >> (problem->x + i)->values[j];
			}
		}
		ifile.close();
	}
	else
	{
		std::cout << "Unable to open point file '" << filename << "'" << std::endl;
	}

	return num_rows;
}
*/

void SVGM::InitialiseSVM()
{
	svm_parameters_.gamma = gamma_;
	svm_parameters_.nu = nu_;
}

void SVGM::PrepareSVMProblem(vnl_matrix<double>& pointset, int num_points, int num_dimensions, struct svm_problem* problem)
{
	// Copy to SVM problem
	problem->l = num_points;
	problem->y = new double[problem->l];
	problem->x = new struct svm_node[problem->l];

	for(int i = 0; i < problem->l; i++)
	{
		problem->y[i] = 1; // Class Label
		(problem->x + i)->dim = num_dimensions; // Dimension
		(problem->x + i)->values = new double[num_dimensions];
		for(int j = 0; j < num_dimensions; ++j)
		{
			(problem->x + i)->values[j] = pointset[i][j];
		}
	}
}

/*
 * Train One-Class SVM
 */
bool SVGM::TrainSVM(const struct svm_problem* problem, struct svm_model** model)
{
	bool ret;

	if(CheckSVMParameters(problem))
	{
		(*model) = svm_train(problem, &svm_parameters_);
		ret = true;
	}
	else
	{
		ret = false;
	}
	return ret;
}

bool SVGM::CheckSVMParameters(const struct svm_problem* problem)
{
	bool ret;
	const char *error_message;

	error_message = svm_check_parameter(problem, &svm_parameters_);

	if(error_message)
	{
		std::cout << "SVM ERROR: " << error_message << std::endl;
		ret = false;
	}
	else
	{
		ret = true;
	}

	return ret;
}

void SVGM::MapToGMMs()
{
	int i, j;
	double t;

	num_components_x_ = model_x_->l;
	num_components_y_ = model_y_->l;

	t = pow(2 * M_PI * sigma_ * sigma_, dimension_ / 2);

	mu_x_.set_size(num_components_x_, dimension_);
	phi_x_.set_size(num_components_x_);
	for(i = 0; i < num_components_x_; ++i)
	{
		for(j = 0; j < dimension_; ++j)
		{
			mu_x_[i][j] = model_x_->SV[i].values[j];
		}
		phi_x_[i] = t * model_x_->sv_coef[0][i];
	}

	mu_y_.set_size(num_components_y_, dimension_);
	phi_y_.set_size(num_components_y_);
	for(i = 0; i < num_components_y_; ++i)
	{
		for(j = 0; j < dimension_; ++j)
		{
			mu_y_[i][j] = model_y_->SV[i].values[j];
		}
		phi_y_[i] = t * model_y_->sv_coef[0][i];
	}
}
