/*
 * Support Vector Registration (SVR): svgm
 * A class for the Support Vector-parametrised Gaussian Mixture (SVGM)
 * data representation. Includes methods for loading point-sets, constructing
 * the SVGM and manipulation.
 * 
 * Campbell, D., and Petersson, L., “An Adaptive Data Representations for
 * Robust Point-Set Registration and Merging”, International Conference on
 * Computer Vision (ICCV), Santiago, Chile, IEEE, Dec. 2015
 * 
 * For the full license, see license.txt in the root directory
 * 
 * Author: Dylan Campbell
 * Date: 20150820
 * Revision: 1
 */

#ifndef SVGM_H_
#define SVGM_H_

#include <iostream>
#include <fstream>
#include <time.h>
#include <vector>
#include <algorithm> // random_shuffle, count
//#include <iterator> // istream_iterator
#include <vnl/vnl_vector.h>
#include <vnl/vnl_matrix.h>

#include "svm.h"

class SVGM
{
public:
	// Constructors
	SVGM();

	// Destructors
	~SVGM();

	// Accessors
	int num_points_x() const {return num_points_x_;}
	int num_points_y() const {return num_points_y_;}
	int num_components_x() const {return num_components_x_;}
	int num_components_y() const {return num_components_y_;}
	int max_num_points() const {return max_num_points_;}
	int dimension() const {return dimension_;}
	double sigma() const {return sigma_;}
	double gamma() const {return gamma_;}
	double nu() const {return nu_;}
	double* durations() {return durations_;}
	struct svm_problem problem_x() const {return problem_x_;}
	struct svm_problem problem_y() const {return problem_y_;}
	struct svm_parameter svm_parameters() const {return svm_parameters_;}
	struct svm_model* model_x() const {return model_x_;}
	struct svm_model* model_y() const {return model_y_;}
	vnl_vector<double> phi_x() const {return phi_x_;}
	vnl_vector<double> phi_y() const {return phi_y_;}
	vnl_matrix<double> mu_x() const {return mu_x_;}
	vnl_matrix<double> mu_y() const {return mu_y_;}

	// Mutators
	void set_num_points_x(int num_points_x) {num_points_x_ = num_points_x;}
	void set_num_points_y(int num_points_y) {num_points_y_ = num_points_y;}
	void set_num_components_x(int num_components_x) {num_components_x_ = num_components_x;}
	void set_num_components_y(int num_components_y) {num_components_y_ = num_components_y;}
	void set_max_num_points(int max_num_points) {max_num_points_ = max_num_points;}
	void set_dimension(int dimension) {dimension_ = dimension;}
	void set_sigma(double sigma) {sigma_ = sigma; gamma_ = 1 / (2 * pow(sigma_, 2));}
	void set_gamma(double gamma) {gamma_ = gamma; sigma_ = 1 / sqrt(2 * gamma_);}
	void set_nu(double nu) {nu_ = nu;}
	void set_durations(double* durations) {std::copy(durations, durations + 4, durations_);}
	void set_problem_x(struct svm_problem problem_x) {problem_x_ = problem_x;}
	void set_problem_y(struct svm_problem problem_y) {problem_y_ = problem_y;}
	void set_svm_parameters(struct svm_parameter svm_parameters) {svm_parameters_ = svm_parameters;}
	void set_model_x(struct svm_model* model_x) {model_x_ = model_x;}
	void set_model_y(struct svm_model* model_y) {model_y_ = model_y;}
	void set_phi_x(vnl_vector<double>& phi_x) {phi_x_ = phi_x;}
	void set_phi_y(vnl_vector<double>& phi_y) {phi_y_ = phi_y;}
	void set_mu_x(vnl_matrix<double>& mu_x) {mu_x_ = mu_x;}
	void set_mu_y(vnl_matrix<double>& mu_y) {mu_y_ = mu_y;}

	// Public Class Functions
	int LoadPointSetX(std::string filename);
	int LoadPointSetY(std::string filename);
	void BuildSVGMs();

private:
	int num_points_x_, num_points_y_; // Number of points in model and scene point-sets
	int num_components_x_, num_components_y_; // Number of components in model and scene GMMs
	int max_num_points_;
	int dimension_; // Point-set dimension
	double sigma_, gamma_, nu_;
	double durations_[3];
	struct svm_problem problem_x_;
	struct svm_problem problem_y_;
	struct svm_parameter svm_parameters_;
	struct svm_model* model_x_;
	struct svm_model* model_y_;
	vnl_vector<double> phi_x_, phi_y_;
	vnl_matrix<double> mu_x_, mu_y_;

	// Private Class Functions
	void InitialiseSVMParameters();
	int LoadPointSet(std::string filename, struct svm_problem* problem);
	void InitialiseSVM();
	void PrepareSVMProblem(vnl_matrix<double>& pointset, int num_points, int num_dimensions, struct svm_problem* problem);
	bool TrainSVM(const struct svm_problem* problem, struct svm_model** model);
	bool CheckSVMParameters(const struct svm_problem* problem);
	void MapToGMMs();
};

#endif /* SVGM_H_ */
