/*
 * Support Vector Registration (SVR): svr
 * An algorithm for robust point-set registration using the Support Vector-
 * parametrised Gaussian Mixture (SVGM) data representation.
 * 
 * Campbell, D., and Petersson, L., “An Adaptive Data Representations for
 * Robust Point-Set Registration and Merging”, International Conference on
 * Computer Vision (ICCV), Santiago, Chile, IEEE, Dec. 2015
 * 
 * For the full license, see license.txt in the root directory
 * 
 * Author: Dylan Campbell
 * Date: 20150820
 * Revision: 1
 */

#include "svr.h"

SVR::SVR()
{
	num_scales_ = -1;
	scales_ = NULL;
//	durations_;
//	num_bounds_;
//	lower_bounds_, upper_bounds_;
//	theta_;
//	mu_y_transformed_;
//	rotation_;
//	translation_;
//	svgm_;
	svr_cost_function_ = new SVRRigidCostFunction;
}

SVR::~SVR()
{
	delete svr_cost_function_;
	delete[] scales_;
}

/*
 * Public Class Functions
 */

void SVR::SetSVGMParameters(int max_num_points, double sigma, double nu)
{
	svgm_.set_max_num_points(max_num_points);
	svgm_.set_sigma(sigma);
	svgm_.set_nu(nu);
}

int SVR::LoadPointSetX(std::string filename)
{
	return svgm_.LoadPointSetX(filename);
}

int SVR::LoadPointSetY(std::string filename)
{
	return svgm_.LoadPointSetY(filename);
}

void SVR::Run()
{
	clock_t  clock_begin;

	svgm_.BuildSVGMs();

	// Add timing information to SVR timing array
	durations_[0] = svgm_.durations()[0]; // TrainSVM X
	durations_[1] = svgm_.durations()[1]; // TrainSVM Y
	durations_[2] = svgm_.durations()[2]; // MapToGMMs

	// Gaussian Mixture Alignment (L2 Distance Minimisation)
	clock_begin = clock();
	StartOptimisation();
	durations_[3] = (double)(clock() - clock_begin) / CLOCKS_PER_SEC;
}

void SVR::TransformMuY(const vnl_vector<double>& theta)
{
	vnl_matrix<double> ones;
	ones.set_size(svgm_.num_components_y(), 1);
	ones.fill(1);

	if(svgm_.dimension() == 2)
	{
		rotation_[0][0] = cos(theta[2]);
		rotation_[0][1] = -sin(theta[2]);
		rotation_[1][0] = sin(theta[2]);
		rotation_[1][1] = cos(theta[2]);
		translation_[0][0] = theta[0];
		translation_[0][1] = theta[1];
	}
	else if(svgm_.dimension() == 3)
	{
		vnl_vector<double> q;
		q.set_size(4);
		for(int i = 0; i < 4; ++i)
		{
			q[i] = theta[i];
		}
		QuaternionToRotation(q, rotation_);
		translation_[0][0] = theta[4];
		translation_[0][1] = theta[5];
		translation_[0][2] = theta[6];
	}

	mu_y_transformed_ = svgm_.mu_y() * rotation_.transpose() + ones * translation_;

	theta_ = theta;
}

double SVR::GaussTransform(double sigma, vnl_matrix<double>& gradient)
{
	return GaussTransform(sigma, gradient.data_block());
}

void SVR::QuaternionToRotation(vnl_vector<double> q, vnl_matrix<double>& R)
{
	// Using: http://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation#Conversion_to_and_from_the_matrix_representation
	// q = a + bi + cj + dk

	double a, b, c, d, a2, b2, c2, d2, bc, cd, bd, ab, ac, ad, z;

	a = q(3);
	b = q(0);
	c = q(1);
	d = q(2);

	a2 = a * a;
	b2 = b * b;
	c2 = c * c;
	d2 = d * d;

	bc = b * c;
	cd = c * d;
	bd = b * d;
	ab = a * b;
	ac = a * c;
	ad = a * d;

	// Diagonal terms
	R(0,0) = a2 + b2 - c2 - d2;
	R(1,1) = a2 - b2 + c2 - d2;
	R(2,2) = a2 - b2 - c2 + d2;

	// Off-diagonal terms
	R(0,1) = 2 * (bc - ad);
	R(0,2) = 2 * (bd + ac);
	R(1,0) = 2 * (bc + ad);
	R(1,2) = 2 * (cd - ab);
	R(2,0) = 2 * (bd - ac);
	R(2,1) = 2 * (cd + ab);

	// Normalise (make rotation matrix orthogonal)
	z = a2 + b2 + c2 + d2;
	R = R / z;
}

void SVR::QuaternionToRotation(vnl_vector<double> q, vnl_matrix<double>& R, vnl_matrix<double>& dR1, vnl_matrix<double>& dR2, vnl_matrix<double>& dR3, vnl_matrix<double>& dR4)
{
	// Using: http://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation#Conversion_to_and_from_the_matrix_representation
	// q = a + bi + cj + dk

	double a, b, c, d, a2, b2, c2, d2, bc, cd, bd, ab, ac, ad, z, z2;

	a = q(3);
	b = q(0);
	c = q(1);
	d = q(2);

	a2 = a * a;
	b2 = b * b;
	c2 = c * c;
	d2 = d * d;

	bc = b * c;
	cd = c * d;
	bd = b * d;
	ab = a * b;
	ac = a * c;
	ad = a * d;

	// Diagonal terms
	R(0,0) = a2 + b2 - c2 - d2;
	R(1,1) = a2 - b2 + c2 - d2;
	R(2,2) = a2 - b2 - c2 + d2;

	// Off-diagonal terms
	R(0,1) = 2 * (bc - ad);
	R(0,2) = 2 * (bd + ac);
	R(1,0) = 2 * (bc + ad);
	R(1,2) = 2 * (cd - ab);
	R(2,0) = 2 * (bd - ac);
	R(2,1) = 2 * (cd + ab);

	// Normalise (make rotation matrix orthogonal)
	z = a2 + b2 + c2 + d2;
	R = R / z;

	// Find the derivative of the rotation matrix with respect to each quaternion parameter
  z2 = z * z;

  // Diagonal terms
  // Derivative of R(1,1) = (a2 + b2 - c2 - d2) / (a2 + b2 + c2 + d2)
  dR1(0,0) = +4 * b * (c2 + d2) / z2;
  dR2(0,0) = -4 * c * (b2 + a2) / z2;
  dR3(0,0) = -4 * d * (b2 + a2) / z2;
  dR4(0,0) = +4 * a * (c2 + d2) / z2;
  // Derivative of R(2,2) = (a2 - b2 + c2 - d2) / (a2 + b2 + c2 + d2)
  dR1(1,1) = -4 * b * (c2 + a2) / z2;
  dR2(1,1) = +4 * c * (b2 + d2) / z2;
  dR3(1,1) = -4 * d * (c2 + a2) / z2;
  dR4(1,1) = +4 * a * (b2 + d2) / z2;
  // Derivative of R(3,3) = (a2 - b2 - c2 + d2) / (a2 + b2 + c2 + d2)
  dR1(2,2) = -4 * b * (d2 + a2) / z2;
  dR2(2,2) = -4 * c * (a2 + d2) / z2;
  dR3(2,2) = +4 * d * (b2 + c2) / z2;
  dR4(2,2) = +4 * a * (b2 + c2) / z2;

  // Off-diagonal terms
  // Derivative of R(0,1) = (2 * (bc - ad)) / (a2 + b2 + c2 + d2)
  dR1(0,1) = +2 * c / z - 2 * b * R(0,1) / z2;
  dR2(0,1) = +2 * b / z - 2 * c * R(0,1) / z2;
  dR3(0,1) = -2 * a / z - 2 * d * R(0,1) / z2;
  dR4(0,1) = -2 * d / z - 2 * a * R(0,1) / z2;
  // Derivative of R(0,2) = (2 * (bd + ac)) / (a2 + b2 + c2 + d2)
  dR1(0,2) = +2 * d / z - 2 * b * R(0,2) / z2;
  dR2(0,2) = +2 * a / z - 2 * c * R(0,2) / z2;
  dR3(0,2) = +2 * b / z - 2 * d * R(0,2) / z2;
  dR4(0,2) = +2 * c / z - 2 * a * R(0,2) / z2;
  // Derivative of R(1,0) = (2 * (bc + ad)) / (a2 + b2 + c2 + d2)
  dR1(1,0) = +2 * c / z - 2 * b * R(1,0) / z2;
  dR2(1,0) = +2 * b / z - 2 * c * R(1,0) / z2;
  dR3(1,0) = +2 * a / z - 2 * d * R(1,0) / z2;
  dR4(1,0) = +2 * d / z - 2 * a * R(1,0) / z2;
  // Derivative of R(1,2) = (2 * (cd - ab)) / (a2 + b2 + c2 + d2)
  dR1(1,2) = -2 * a / z - 2 * b * R(1,2) / z2;
  dR2(1,2) = +2 * d / z - 2 * c * R(1,2) / z2;
  dR3(1,2) = +2 * c / z - 2 * d * R(1,2) / z2;
  dR4(1,2) = -2 * b / z - 2 * a * R(1,2) / z2;
  // Derivative of R(2,0) = (2 * (bd - ac)) / (a2 + b2 + c2 + d2)
  dR1(2,0) = +2 * d / z - 2 * b * R(2,0) / z2;
  dR2(2,0) = -2 * a / z - 2 * c * R(2,0) / z2;
  dR3(2,0) = +2 * b / z - 2 * d * R(2,0) / z2;
  dR4(2,0) = -2 * c / z - 2 * a * R(2,0) / z2;
  // Derivative of R(2,1) = (2 * (cd + ab)) / (a2 + b2 + c2 + d2)
  dR1(2,1) = +2 * a / z - 2 * b * R(2,1) / z2;
  dR2(2,1) = +2 * d / z - 2 * c * R(2,1) / z2;
  dR3(2,1) = +2 * c / z - 2 * d * R(2,1) / z2;
  dR4(2,1) = +2 * b / z - 2 * a * R(2,1) / z2;
}

/*
 * Private Class Functions
 */

void SVR::StartOptimisation()
{
	int i;
	double sigma;
	vnl_lbfgsb solver(*svr_cost_function_);

	SetBounds();
	solver.set_bound_selection(num_bounds_);
	solver.set_lower_bound(lower_bounds_);
	solver.set_upper_bound(upper_bounds_);
	SetOptimisationOptions(solver);

	svr_cost_function_->Initialise(this, &svgm_);

	SetInitialTheta();
	sigma = svgm_.sigma();

	// Convert SVGM sigma to SVR sigma
	// In the paper, sqrt(2) was used for all experiments
	// Higher values may improve accuracy, but may decrease stability
	// ToDo: make this constant a config file parameter
	sigma /= sqrt(2.0);

	// Annealing disabled
	if(num_scales_ == 0)
	{
		svr_cost_function_->set_sigma(sigma);
		if(!solver.minimize(theta_))
		{
			std::cout << "Solver failed with error code: " << solver.get_failure_code() << std::endl;
		}
	}
	// Annealing enabled
	else
	{
		for(i = 0; i < num_scales_; ++i)
		{
			sigma = sigma * scales_[i];

			// To retrain SVMs when annealing
			if(i > 0)
			{
				svgm_.set_sigma(sigma); // Also sets gamma
				svgm_.BuildSVGMs();
				svr_cost_function_->Initialise(this, &svgm_);
			}

			svr_cost_function_->set_sigma(sigma);
			if(!solver.minimize(theta_))
			{
				std::cout << "Solver failed with error code: " << solver.get_failure_code() << std::endl;
				break;
			}

//			std::cout << std::endl
//								<< "OptVal:      " << svr_cost_function_->f(theta_) << std::endl
//								<< "Iterations:  " << solver.get_num_iterations() << std::endl
//								<< "Evaluations: " << solver.get_num_evaluations() << std::endl;
		}
	}
}

void SVR::SetBounds()
{
	if(svgm_.dimension() == 2)
	{
		num_bounds_.set_size(3); // [x y alpha]
		num_bounds_.fill(0); // unconstrained
		num_bounds_[2] = 2; // constrained with lower and upper bounds
		lower_bounds_.set_size(3);
		lower_bounds_.fill(0);
		lower_bounds_[2] = -M_PI;
		upper_bounds_.set_size(3);
		upper_bounds_.fill(0);
		upper_bounds_[2] = -M_PI;
	}
	else if(svgm_.dimension() == 3)
	{
		num_bounds_.set_size(7); // [qx qy qz qw tx ty tz]
		num_bounds_.fill(0); // unconstrained
		for(int i = 0; i < 4; ++i)
		{
			num_bounds_[i] = 2; // constrained with lower and upper bounds
		}
		lower_bounds_.set_size(7);
		lower_bounds_.fill(0);
		for(int i = 0; i < 4; ++i)
		{
			lower_bounds_[i] = -1;
		}
		upper_bounds_.set_size(7);
		upper_bounds_.fill(0);
		for(int i = 0; i < 4; ++i)
		{
			upper_bounds_[i] = 1;
		}
	}
}

/*
 * Initialise theta to the identity transformation and set the size of the rotation and translation matrices
 */
void SVR::SetInitialTheta()
{
	if(svgm_.dimension() == 2)
	{
		theta_.set_size(3);
		theta_.fill(0);
		rotation_.set_size(2, 2);
		translation_.set_size(1, 2);
	}
	else if(svgm_.dimension() == 3)
	{
		theta_.set_size(7);
		theta_.fill(0);
		theta_[3] = 1;
		rotation_.set_size(3, 3);
		translation_.set_size(1, 3);
	}
}

void SVR::SetOptimisationOptions(vnl_lbfgsb& solver)
{
	// Set the convergence tolerance on F (sum of squared residuals).
	// When the differences in successive RMS errors is less than this, the
	// routine terminates.  So this is effectively the desired precision of your
	// minimization.  Setting it too low wastes time, too high might cause early
	// convergence.  The default of 1e-9 is on the safe side, but if speed is an
	// issue, you can try raising it.
	solver.set_f_tolerance(1e-9);

	// Set the convergence tolerance on X.
	// When the length of the steps taken in X are about this long, the routine
	// terminates.  The default is 1e-8, which should work for many problems,
	// but if you can get away with 1e-4, say, minimizations will be much quicker.
	solver.set_x_tolerance(1e-8);

	// Set the convergence tolerance on Grad(F)' * F.
	// The default value is 1e-5
	solver.set_g_tolerance(1e-5);

	// Set the termination maximum number of iterations.
	solver.set_max_function_evals(10000);

	// Set the cost function convergence factor.
	// When an iteration changes the function value by an amount smaller than
	// this factor times the machine epsilon (scaled by function magnitude)
	// convergence is assumed.  The default value is 1e+7.
	solver.set_cost_function_convergence_factor(1e3);

	// Set the projected gradient tolerance.
	// When the projected gradient vector has no component larger than
	// the given value convergence is assumed.  The default value is 1e-5.
	solver.set_projected_gradient_tolerance(1e-5);
}

double SVR::GaussTransform(const double* x, const double* y, const double* phi_x, const double* phi_y, const double sigma, double* gradient)
{
	int i, j, k;
	int ik, jk;
	double dist2_ij;
	double cost_ij;
	double fval;
	double c;

	// Initialise variables
	c = 1e8; // avoids numerical issues with optimisation
	fval = 0;
	for(i = 0; i < svgm_.num_components_y() * svgm_.dimension(); ++i)
	{
		gradient[i] = 0;
	}

	// For all points x in X
	for(i = 0; i < svgm_.num_components_x(); ++i)
	{
		// For all points y in Y
		for(j = 0; j < svgm_.num_components_y(); ++j)
		{
			// Calculate the squared distance between points x and y
			dist2_ij = 0;
			for(k = 0; k < svgm_.dimension(); ++k)
			{
				ik = i * svgm_.dimension() + k;
				jk = j * svgm_.dimension() + k;
				dist2_ij += (x[ik] - y[jk]) * (x[ik] - y[jk]);
			}

			// Evaluate the pointwise summand of the objective function
			cost_ij = c * phi_x[i] * phi_y[j] * exp(-dist2_ij / (4 * sigma * sigma));

			// Evaluate the gradient
			for(k = 0; k < svgm_.dimension(); ++k)
			{
				ik = i * svgm_.dimension() + k;
				jk = j * svgm_.dimension() + k;
				gradient[jk] += -cost_ij * (y[jk] - x[ik]);
			}

			fval += cost_ij;
		}
	}

	for(i = 0; i < (svgm_.num_components_y() * svgm_.dimension()); ++i)
	{
		gradient[i] /= (svgm_.num_components_x() * svgm_.num_components_y() * sigma * sigma);
	}

	return fval / (svgm_.num_components_x() * svgm_.num_components_y());
}

double SVR::GaussTransform(const double sigma, double* gradient)
{
	return GaussTransform(svgm_.mu_x().data_block(), mu_y_transformed_.data_block(), svgm_.phi_x().data_block(), svgm_.phi_y().data_block(), sigma, gradient);
}
