/*
 * Support Vector Registration (SVR): svr
 * An algorithm for robust point-set registration using the Support Vector-
 * parametrised Gaussian Mixture (SVGM) data representation.
 * 
 * Campbell, D., and Petersson, L., “An Adaptive Data Representations for
 * Robust Point-Set Registration and Merging”, International Conference on
 * Computer Vision (ICCV), Santiago, Chile, IEEE, Dec. 2015
 * 
 * For the full license, see license.txt in the root directory
 * 
 * Author: Dylan Campbell
 * Date: 20150820
 * Revision: 1
 */

#ifndef SVR_H_
#define SVR_H_

#include <iostream>
#include <fstream>
#include <time.h>
//#include <vector>
//#include <algorithm> // count
//#include <iterator> // istream_iterator
#include <vnl/vnl_vector.h>
#include <vnl/vnl_matrix.h>
#include <vnl/algo/vnl_lbfgsb.h>

#include "svgm.h"
#include "svr_rigid_cost_function.h"

class SVRRigidCostFunction; // Forward declare class SVRRigidCostFunction

class SVR
{
public:
	// Constructors
	SVR();

	// Destructors
	~SVR();

	// Accessors
	int num_scales() const {return num_scales_;}
	double* sigmas() const {return scales_;}
	double* durations() {return durations_;}
	vnl_vector<long> num_bounds() const {return num_bounds_;}
	vnl_vector<double> lower_bounds() const {return lower_bounds_;}
	vnl_vector<double> upper_bounds() const {return upper_bounds_;}
	vnl_vector<double> theta() const {return theta_;}
	vnl_matrix<double> mu_y_transformed() const {return mu_y_transformed_;}
	vnl_matrix<double> rotation() const {return rotation_;}
	vnl_matrix<double> translation() const {return translation_;}
	SVGM svgm() const {return svgm_;}
	SVRRigidCostFunction* svr_cost_function() const {return svr_cost_function_;}

	// Mutators
	void set_num_scales(int num_scales) {num_scales_ = num_scales;}
	void set_scales(float* scales) {scales_ = new double[num_scales_]; std::copy(scales, scales + num_scales_, scales_);}
	void set_durations(double* durations) {std::copy(durations, durations + 4, durations_);}
	void set_num_bounds(vnl_vector<long>& num_bounds) {num_bounds_ = num_bounds;}
	void set_lower_bounds(vnl_vector<double>& lower_bounds) {lower_bounds_ = lower_bounds;}
	void set_upper_bounds(vnl_vector<double>& upper_bounds) {upper_bounds_ = upper_bounds;}
	void set_theta(vnl_vector<double>& theta) {theta_ = theta;}
	void set_mu_y_transformed(vnl_matrix<double>& mu_y_transformed) {mu_y_transformed_ = mu_y_transformed;}
	void set_rotation(vnl_matrix<double>& rotation) {rotation_ = rotation;}
	void set_translation(vnl_matrix<double>& translation) {translation_ = translation;}
	void set_svgm(SVGM svgm) {svgm_ = svgm;}
	void set_svr_cost_function(SVRRigidCostFunction* svr_cost_function) {svr_cost_function_ = svr_cost_function;}

	// Public Class Functions
	void SetSVGMParameters(int max_num_points, double sigma, double nu);
	int LoadPointSetX(std::string filename);
	int LoadPointSetY(std::string filename);
	void Run();
	void TransformMuY(const vnl_vector<double>& theta);
	double GaussTransform(const double sigma, vnl_matrix<double>& gradient);
	void QuaternionToRotation(vnl_vector<double> q, vnl_matrix<double>& R);
	void QuaternionToRotation(vnl_vector<double> q, vnl_matrix<double>& R, vnl_matrix<double>& dR1, vnl_matrix<double>& dR2, vnl_matrix<double>& dR3, vnl_matrix<double>& dR4);

private:
	int num_scales_;
	double* scales_;
	double durations_[4];
	vnl_vector<long> num_bounds_;
	vnl_vector<double> lower_bounds_, upper_bounds_;
	vnl_vector<double> theta_;
	vnl_matrix<double> mu_y_transformed_;
	vnl_matrix<double> rotation_;
	vnl_matrix<double> translation_;
	SVGM svgm_;
	SVRRigidCostFunction* svr_cost_function_;

	// Private Class Functions
	void StartOptimisation();
	void SetBounds();
	void SetInitialTheta();
	void SetOptimisationOptions(vnl_lbfgsb& solver);
	double GaussTransform(const double* x, const double* y, const double* phi_x, const double* phi_y, const double sigma, double* gradient);
	double GaussTransform(const double sigma, double* gradient);
};

#endif /* SVR_H_ */
