/*
 * Support Vector Registration (SVR): svr_api
 * API for the SVR algorithm.
 * 
 * Campbell, D., and Petersson, L., “An Adaptive Data Representations for
 * Robust Point-Set Registration and Merging”, International Conference on
 * Computer Vision (ICCV), Santiago, Chile, IEEE, Dec. 2015
 * 
 * For the full license, see license.txt in the root directory
 * 
 * Author: Dylan Campbell
 * Date: 20150820
 * Revision: 1
 */

#include <time.h>
#include <iostream>
#include <fstream>

#include "svr.h"
#include "config_map.h"

#define DEFAULT_OUTPUT_FNAME "output.txt"
#define DEFAULT_CONFIG_FNAME "config.txt"
#define DEFAULT_POINTSET_X_FNAME "x.txt"
#define DEFAULT_POINTSET_Y_FNAME "y.txt"

using namespace std;

void ParseInput(int argc, char **argv, string & pointset_x_filename, string & pointset_y_filename, string & config_filename, string & output_filename);
void ReadConfig(string filename, SVR & svr);

int main(int argc, char* argv[])
{
	int num_points_x, num_points_y;
	double total_duration;
	ofstream ofile;
	clock_t  clock_begin, clock_end;
	string pointset_x_filename, pointset_y_filename, config_filename, output_filename;
	SVR svr;

	ParseInput(argc, argv, pointset_x_filename, pointset_y_filename, config_filename, output_filename);
	ReadConfig(config_filename, svr);

	// Load model and data point clouds (outside Run() function since it is file I/O)
	num_points_x = svr.LoadPointSetX(pointset_x_filename);
	num_points_y = svr.LoadPointSetY(pointset_y_filename);
	
	// Run SVR
	cout << "Point-Set X ID: " << pointset_x_filename << " (" << num_points_x << "), Point-Set Y ID: " << pointset_y_filename << " (" << num_points_y << ")" << endl;
	clock_begin = clock();
	svr.Run();
	clock_end = clock();
	
	// Outputs
	total_duration = (double)(clock_end - clock_begin) / CLOCKS_PER_SEC;

	cout << "Theta [rotation quaternion | translation vector]:" << endl;
	cout << svr.theta() << endl << endl;
	cout << "Rotation Matrix:" << endl;
	cout << svr.rotation() << endl;
	cout << "Translation Vector:" << endl;
	cout << svr.translation() << endl;
	cout << "Duration: " << total_duration << "s" << endl;
	cout << "Training SVMs Took: " << svr.durations()[0] + svr.durations()[1] << "s" << endl;
	cout << "Mapping to GMMs Took: " << svr.durations()[2] << "s" << endl;
	cout << "Optimisation Took: " << svr.durations()[3] << "s" << endl;

	ofile.open(output_filename.c_str(), ofstream::out);
	ofile << svr.theta() << endl;
	ofile << svr.durations()[0] << " " << svr.durations()[1] << " " << svr.durations()[3] << endl;
//	ofile << total_duration << endl;
//	ofile << svr.rotation() << endl;
//	ofile << svr.translation() << endl;
	ofile.close();

	return 0;
}

void ParseInput(int argc, char **argv, string & pointset_x_filename, string & pointset_y_filename, string & config_filename, string & output_filename)
{
	// Set default values
	pointset_x_filename = DEFAULT_POINTSET_X_FNAME;
	pointset_y_filename = DEFAULT_POINTSET_Y_FNAME;
	config_filename = DEFAULT_CONFIG_FNAME;
	output_filename = DEFAULT_OUTPUT_FNAME;

	cout << "Support Vector Registration" << endl;
	cout << "Copyright (C) 2015 Dylan Campbell & Lars Petersson" << endl;
	cout << "USAGE:" << argv[0] << " <POINT-SET X FILENAME> <POINT-SET Y FILENAME> <CONFIG FILENAME> <OUTPUT FILENAME>" << endl;
	cout << "OUTPUT: [rotation quaternion | translation vector] transforming point-set Y to align with point-set X" << endl;
	cout << endl;

	if(argc > 4)
	{
		output_filename = argv[4];
	}
	if(argc > 3)
	{
		config_filename = argv[3];
	}
	if(argc > 2)
	{
		pointset_y_filename = argv[2];
	}
	if(argc > 1)
	{
		pointset_x_filename = argv[1];
	}

	cout << "INPUT:" << endl;
	cout << "(pointset_x_filename)->(" << pointset_x_filename << ")" << endl;
	cout << "(pointset_y_filename)->(" << pointset_y_filename << ")" << endl;
	cout << "(config_filename)->(" << config_filename << ")" << endl;
	cout << "(output_filename)->(" << output_filename << ")" << endl;
	cout << endl;
}

void ReadConfig(string filename, SVR & svr)
{
	// Open and parse the associated config file
	ConfigMap config(filename.c_str());

	svr.set_num_scales(config.getI("num_scales"));
	svr.set_scales(config.getVector("scales"));
	svr.SetSVGMParameters(config.getI("max_num_points"), config.getF("sigma"), config.getF("nu"));

	cout << "CONFIG:" << endl;
	config.print();
	cout << endl;
}
