/*
 * Support Vector Registration (SVR): svr_rigid_cost_function
 * L2 cost function for rigid registration. Required for VNL LBFGSB optimiser.
 * 
 * Campbell, D., and Petersson, L., “An Adaptive Data Representations for
 * Robust Point-Set Registration and Merging”, International Conference on
 * Computer Vision (ICCV), Santiago, Chile, IEEE, Dec. 2015
 * 
 * For the full license, see license.txt in the root directory
 * 
 * Author: Dylan Campbell
 * Date: 20150820
 * Revision: 1
 */

#include "svr_rigid_cost_function.h"

SVRRigidCostFunction::SVRRigidCostFunction() : vnl_cost_function()
{
  dimension_ = 0;
  sigma_ = 0.01;
//  gradient_;
  svr_ = NULL;
}

double SVRRigidCostFunction::f(const vnl_vector<double>& theta)
{
	double fval;
	svr_->TransformMuY(theta);
	fval = svr_->GaussTransform(sigma_, gradient_);
	fval = -fval;
	gradient_ = -gradient_;
	return fval;
}

void SVRRigidCostFunction::gradf(const vnl_vector<double>& theta, vnl_vector<double>& fgrad)
{
	int i, j;
	vnl_matrix<double> GmuY;

	if(dimension_ == 2)
	{
		fgrad[0] = gradient_.get_column(0).sum(); // Eq 16
		fgrad[1] = gradient_.get_column(1).sum(); // Eq 16

		vnl_matrix<double> dR;
		dR.set_size(2, 2);
		dR[0][0] = -sin(theta[2]);
		dR[0][1] = -cos(theta[2]);
		dR[1][0] = +cos(theta[2]);
		dR[1][1] = -sin(theta[2]);

		GmuY = gradient_.transpose() * svgm_->mu_y();

		fgrad[2] = 0;
		for(i = 0; i < 2; ++i)
		{
			for(j = 0; j < 2; ++j)
			{
				fgrad[2] += GmuY[i][j] * dR[i][j]; // Eq 17
			}
		}
	}
	else if(dimension_ == 3)
	{
		fgrad[4] = gradient_.get_column(0).sum(); // Eq 16
		fgrad[5] = gradient_.get_column(1).sum(); // Eq 16
		fgrad[6] = gradient_.get_column(2).sum(); // Eq 16

		vnl_vector<double> q;
		q.set_size(4);
		for(i = 0; i < 4; ++i)
		{
			q[i] = theta[i];
		}
		vnl_matrix<double> R;
		R.set_size(3, 3);
		vnl_matrix<double> dR1, dR2, dR3, dR4;
		dR1.set_size(3, 3);
		dR2.set_size(3, 3);
		dR3.set_size(3, 3);
		dR4.set_size(3, 3);
		svr_->QuaternionToRotation(q, R, dR1, dR2, dR3, dR4);

		GmuY = gradient_.transpose() * svgm_->mu_y();

		fgrad[0] = 0;
		for(i = 0; i < 3; ++i)
		{
			for(j = 0; j < 3; ++j)
			{
				fgrad[0] += GmuY[i][j] * dR1[i][j]; // Eq 17
			}
		}
		fgrad[1] = 0;
		for(i = 0; i < 3; ++i)
		{
			for(j = 0; j < 3; ++j)
			{
				fgrad[1] += GmuY[i][j] * dR2[i][j]; // Eq 17
			}
		}
		fgrad[2] = 0;
		for(i = 0; i < 3; ++i)
		{
			for(j = 0; j < 3; ++j)
			{
				fgrad[2] += GmuY[i][j] * dR3[i][j]; // Eq 17
			}
		}
		fgrad[3] = 0;
		for(i = 0; i < 3; ++i)
		{
			for(j = 0; j < 3; ++j)
			{
				fgrad[3] += GmuY[i][j] * dR4[i][j]; // Eq 17
			}
		}
	}
}

void SVRRigidCostFunction::Initialise(SVR* svr, SVGM* svgm)
{
	set_svr(svr);
	set_svgm(svgm);
	dimension_ = svgm_->dimension();
	gradient_.set_size(svgm_->num_components_y(), dimension_);

	if(dimension_ == 2)
	{
		set_number_of_unknowns(3); // [x y alpha]
	}
	else if(dimension_ == 3)
	{
		set_number_of_unknowns(7); // [qx qy qz qw tx ty tz]
	}
}
