%% Support Vector Registration (SVR): config2d
% Configuration file for the SVR algorithm
%
% Campbell, D., and Petersson, L., “An Adaptive Data Representations for
% Robust Point-Set Registration and Merging”, International Conference on
% Computer Vision (ICCV), Santiago, Chile, IEEE, Dec. 2015
%
% For the full license, see license.txt in the root directory
%
% Author: Dylan Campbell
% Date: 20150709
% Revision: 5

function [p] = config2d()

    % Add relevant folders to the path
    addpath(genpath(pwd));
    libsvmPath = '/usr/local/src/libsvm/matlab';
    addpath(libsvmPath);

    % Transform type: {'rigid2d', 'rigid3d', 'affine2d', 'affine3d'}
    p.transformType = 'rigid2d';

    %% One-class SVM parameters
    
    % Maximum number of (randomly selected) points with which to train the SVM
    % Set to Inf to use all points
    p.nSVR = Inf;

    % Nu parameter for SVM: lower bound on fraction of support vectors, upper
    % bound on the misclassification rate
    p.nu = 0.01;

    % Scale parameter for SVM
    p.sigma = 0.1;
    % Sets gamma from sigma
    p.gamma = 1 / (2 * power(p.sigma, 2));
    % Estimate scale from the point-sets instead of using the provided value
    p.estimateSigma = true;

    %% Optimisation Parameters
    
    % Set factor by which optimisation sigma differs from SVGM sigma
    % SVR sigma = SVGM sigma / p.sigmaFactor
    % Set to sqrt(2) for experiments in the paper
    % A larger value can often produce a more refined registration result
    % but can be less stable
    p.sigmaFactor = 1.0;
    
    % Maximum number of optimisation iterations
    p.maxIter = 1000;

    % Constraint on rotation parameters: {'reproject', 'unitnorm'}
    % reproject is used for experiments in the paper
    % unitnorm is only available for rigid3d
    % reproject tends to be more stable, unitnorm is faster
    p.constraint = 'reproject';

    % Sets transformation parameters initial values and bounds
    switch lower(p.transformType)
        case 'rigid2d'
            % [tx ty r]
            p.initParam = [0, 0, 0];
            p.lb = [-inf; -inf; -pi];
            p.ub = [+inf; +inf; +pi];
        case 'rigid3d'
            % [qx qy qz qw tx ty tz]
            p.initParam = [0, 0, 0, 1, 0, 0, 0];
            switch lower(p.constraint)
                case 'reproject'
                    p.lb = [-1; -1; -1; -1; -inf; -inf; -inf];
                    p.ub = [+1; +1; +1; +1; +inf; +inf; +inf];
                case 'unitnorm'
                    p.lb = [-inf; -inf; -inf; -inf; -inf; -inf; -inf];
                    p.ub = [+inf; +inf; +inf; +inf; +inf; +inf; +inf];
                otherwise
                    error('No rotation constraint specified');
            end
        case 'affine2d'
            % [tx ty r11 r12 r21 r22]
            p.initParam = [0, 0, 1, 0, 0, 1];
            p.lb = [-inf; -inf; -inf; -inf; -inf; -inf];
            p.ub = [+inf; +inf; +inf; +inf; +inf; +inf];
        case 'affine3d'
            % [tx ty tz r11 r12 r13 r21 r22 r23 r31 r32 r33]
            p.initParam = [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1];
            p.lb = [-inf; -inf; -inf;    0; -inf; -inf; -inf;    0; -inf; -inf; -inf;    0];
            p.ub = [+inf; +inf; +inf; +inf; +inf; +inf; +inf; +inf; +inf; +inf; +inf; +inf];
        otherwise
            error('Unknown motion type');
    end
end

