// Support Vector Registration (SVR): GaussTransform
// C++ Gauss Transform function with non-uniform weights
// 
// Campbell, D., and Petersson, L., “An Adaptive Data Representations for
// Robust Point-Set Registration and Merging”, International Conference on
// Computer Vision (ICCV), Santiago, Chile, IEEE, Dec. 2015
// 
// For the full license, see license.txt in the root directory
// 
// Author: Dylan Campbell
// Date: 20150709
// Revision: 5
//
// To Build: mex CFLAGS="\$CFLAGS -std=c99" GaussTransformWrapper.c GaussTransform.c -output GaussTransformWrapper

#define SQR(X)  ((X)*(X))

#include <math.h>

#ifdef WIN32
__declspec( dllexport )
#endif

double GaussTransform(const double* X, const double* Y, double* phiX, double* phiY,  int m, int n, int dim, double scale, double* gradient)
{
	int i, j, d; 
    int id, jd;
    double constant;
	double dist2_ij;
    double cost_ij;
    double fval;
    
    // Initialise variables
    constant = 1e11; // avoids numerical issues with optimisation
    fval = 0;
	for(i = 0; i < m * dim; ++i)
    {
        gradient[i] = 0;
    }
    
    // For all points x in X
	for(i = 0; i < m; ++i)
	{
        // For all points y in Y
		for(j = 0; j < n; ++j)
		{
            // Calculate the squared distance between points x and y
			dist2_ij = 0;
			for(d = 0; d < dim; ++d)
			{
                id = i * dim + d;
                jd = j * dim + d;
				dist2_ij += SQR(X[id] - Y[jd]);
			}
            
            // Evaluate the pointwise summand of the objective function
            cost_ij = constant * phiX[i] * phiY[j] * exp(-dist2_ij / (4 * SQR(scale)));
            
            // Evaluate the gradient
            for(d = 0; d < dim; ++d){
                id = i*dim + d;
                jd = j*dim + d;
                gradient[id] += -cost_ij * (X[id] - Y[jd]);
            }
            
			fval += cost_ij;
		}
	}
    
	for(i = 0; i < (m * dim); ++i)
    {
		gradient[i] /= (m * n * SQR(scale));
	}
    
	return fval / (m * n);
}
