%% Support Vector Registration (SVR): quaternionToRotation
% Function to convert a quaternion to a rotation matrix
% and (optionally) compute the gradient
%
% Campbell, D., and Petersson, L., “An Adaptive Data Representations for
% Robust Point-Set Registration and Merging”, International Conference on
% Computer Vision (ICCV), Santiago, Chile, IEEE, Dec. 2015
%
% For the full license, see license.txt in the root directory
%
% Author: Dylan Campbell
% Date: 20150709
% Revision: 5
%
% Input:
%   q: 1 x 4 quaternion [rx, ry, rz, rw]
% Output:
%   R: 3 x 3 rotation matrix
%   dR: 1 x 4 cell array of 3 x 3 derivative matrices

function [R, dR] = quaternionToRotation(q)

% Source: http://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation

% q = a + bi + cj + dk
a = q(4);
b = q(1);
c = q(2);
d = q(3);

a2 = a * a;
b2 = b * b;
c2 = c * c;
d2 = d * d;

bc = b * c;
cd = c * d;
bd = b * d;
ab = a * b;
ac = a * c;
ad = a * d;

% Diagonal terms
R(1,1) = a2 + b2 - c2 - d2;
R(2,2) = a2 - b2 + c2 - d2;
R(3,3) = a2 - b2 - c2 + d2;

% Off-diagonal terms
R(1,2) = 2 * (bc - ad);
R(1,3) = 2 * (bd + ac);
R(2,1) = 2 * (bc + ad);
R(2,3) = 2 * (cd - ab);
R(3,1) = 2 * (bd - ac);
R(3,2) = 2 * (cd + ab);

% Normalise (make rotation matrix orthogonal)
z = a2 + b2 + c2 + d2;
R = R / z;

% If required, find the derivative of the rotation matrix with respect to
% each quaternion parameter
if (nargout > 1)
    
    z2 = z * z;
    
    % Diagonal terms
    % Derivative of R(1,1) = (a2 + b2 - c2 - d2) / (a2 + b2 + c2 + d2)
    dR1(1,1) = +4 * b * (c2 + d2) / z2;
    dR2(1,1) = -4 * c * (b2 + a2) / z2;
    dR3(1,1) = -4 * d * (b2 + a2) / z2;
    dR4(1,1) = +4 * a * (c2 + d2) / z2;
    % Derivative of R(2,2) = (a2 - b2 + c2 - d2) / (a2 + b2 + c2 + d2)
    dR1(2,2) = -4 * b * (c2 + a2) / z2;
    dR2(2,2) = +4 * c * (b2 + d2) / z2;
    dR3(2,2) = -4 * d * (c2 + a2) / z2;
    dR4(2,2) = +4 * a * (b2 + d2) / z2;
    % Derivative of R(3,3) = (a2 - b2 - c2 + d2) / (a2 + b2 + c2 + d2)
    dR1(3,3) = -4 * b * (d2 + a2) / z2;
    dR2(3,3) = -4 * c * (a2 + d2) / z2;
    dR3(3,3) = +4 * d * (b2 + c2) / z2;
    dR4(3,3) = +4 * a * (b2 + c2) / z2;

    % Off-diagonal terms
    % Derivative of R(1,2) = (2 * (bc - ad)) / (a2 + b2 + c2 + d2)
    dR1(1,2) = +2 * c / z - 2 * b * R(1,2) / z2;
    dR2(1,2) = +2 * b / z - 2 * c * R(1,2) / z2;
    dR3(1,2) = -2 * a / z - 2 * d * R(1,2) / z2;
    dR4(1,2) = -2 * d / z - 2 * a * R(1,2) / z2;
    % Derivative of R(1,3) = (2 * (bd + ac)) / (a2 + b2 + c2 + d2)
    dR1(1,3) = +2 * d / z - 2 * b * R(1,3) / z2;
    dR2(1,3) = +2 * a / z - 2 * c * R(1,3) / z2;
    dR3(1,3) = +2 * b / z - 2 * d * R(1,3) / z2;
    dR4(1,3) = +2 * c / z - 2 * a * R(1,3) / z2;
    % Derivative of R(2,1) = (2 * (bc + ad)) / (a2 + b2 + c2 + d2)
    dR1(2,1) = +2 * c / z - 2 * b * R(2,1) / z2;
    dR2(2,1) = +2 * b / z - 2 * c * R(2,1) / z2;
    dR3(2,1) = +2 * a / z - 2 * d * R(2,1) / z2;
    dR4(2,1) = +2 * d / z - 2 * a * R(2,1) / z2;
    % Derivative of R(2,3) = (2 * (cd - ab)) / (a2 + b2 + c2 + d2)
    dR1(2,3) = -2 * a / z - 2 * b * R(2,3) / z2;
    dR2(2,3) = +2 * d / z - 2 * c * R(2,3) / z2;
    dR3(2,3) = +2 * c / z - 2 * d * R(2,3) / z2;
    dR4(2,3) = -2 * b / z - 2 * a * R(2,3) / z2;
    % Derivative of R(3,1) = (2 * (bd - ac)) / (a2 + b2 + c2 + d2)
    dR1(3,1) = +2 * d / z - 2 * b * R(3,1) / z2;
    dR2(3,1) = -2 * a / z - 2 * c * R(3,1) / z2;
    dR3(3,1) = +2 * b / z - 2 * d * R(3,1) / z2;
    dR4(3,1) = -2 * c / z - 2 * a * R(3,1) / z2;
    % Derivative of R(3,2) = (2 * (cd + ab)) / (a2 + b2 + c2 + d2)
    dR1(3,2) = +2 * a / z - 2 * b * R(3,2) / z2;
    dR2(3,2) = +2 * d / z - 2 * c * R(3,2) / z2;
    dR3(3,2) = +2 * c / z - 2 * d * R(3,2) / z2;
    dR4(3,2) = +2 * b / z - 2 * a * R(3,2) / z2;
    
    dR{1} = dR1;
    dR{2} = dR2;
    dR{3} = dR3;
    dR{4} = dR4;
end