%% Support Vector Registration (SVR): svrObjFunc
% L2 objective function for the SVR algorithm
%
% Campbell, D., and Petersson, L., “An Adaptive Data Representations for
% Robust Point-Set Registration and Merging”, International Conference on
% Computer Vision (ICCV), Santiago, Chile, IEEE, Dec. 2015
%
% For the full license, see license.txt in the root directory
%
% Author: Dylan Campbell
% Date: 20150709
% Revision: 5
%
% Input:
%   theta: transformation parameters
%   p: struct of parameters
% Output:
%   fval: function value evaluated with argument theta
%   fgrad: function gradient df/dtheta

function [fval, fgrad] = svrObjFunc(theta, p)

    muX = p.muX;
    muY = p.muY;
    phiX = p.phiX;
    phiY = p.phiY;
    scale = p.sigma;
    transformType = p.transformType;

    [muXt] = transformPointset(muX, transformType, theta);

    switch lower(transformType)
        case 'rigid2d'
            [fval, G] = objFuncRigid(muXt, muY, phiX, phiY, scale); % Eq 14 and 15
            
            r = theta(3);
            dR = [-sin(r) -cos(r); cos(r) -sin(r)]; % dR/dr
            GmuX = G' * muX;
            
            fgrad(1) = sum(G(:, 1)); % Eq 16
            fgrad(2) = sum(G(:, 2)); % Eq 16
            fgrad(3) = sum(sum(GmuX .* dR)); % Eq 17

        case 'rigid3d'
            [fval, G] = objFuncRigid(muXt, muY, phiX, phiY, scale); % Eq 14 and 15
            [~, dR] = quaternionToRotation(theta(1:4)); % dR/dq
            
            GmuX = G' * muX; 
            fgrad(1) = sum(sum(GmuX .* dR{1})); % Eq 17
            fgrad(2) = sum(sum(GmuX .* dR{2})); % Eq 17
            fgrad(3) = sum(sum(GmuX .* dR{3})); % Eq 17
            fgrad(4) = sum(sum(GmuX .* dR{4})); % Eq 17
            fgrad(5) = sum(G(:, 1)); % Eq 16
            fgrad(6) = sum(G(:, 2)); % Eq 16
            fgrad(7) = sum(G(:, 3)); % Eq 16

        case 'affine2d'
            [fval, G] = objFunc(muXt, muY, phiX, phiY, scale);
            
            fgrad(1) = sum(G(:, 1));
            fgrad(2) = sum(G(:, 2));
            fgrad(3:6) = reshape(G' * muX, 1, 4);

        case 'affine3d'
            [fval, G] = objFunc(muXt, muY, phiX, phiY, scale);
            
            fgrad(1) = sum(G(:, 1));
            fgrad(2) = sum(G(:, 2));
            fgrad(3) = sum(G(:, 3));
            fgrad(4:12) = reshape(G' * muX, 1, 9);

        otherwise
            error('Unknown motion type');
    end
end

function [f, g] = objFunc(X, Y, phiX, phiY, scale)
    [f1, g1] = GaussTransformWrapper(X', X', phiX, phiY, scale);
    [f2, g2] = GaussTransformWrapper(X', Y', phiX, phiY, scale);
    f = f1 - 2 * f2;
    g = 2 * g1' - 2 * g2';
end

function [f, g] = objFuncRigid(X, Y, phiX, phiY, scale)
    [f, g] =  GaussTransformWrapper(X', Y', phiX, phiY, scale);
    f = -f;
    g = -g';
end
