%% Support Vector Registration (SVR): transformPointset
% Function to transform a point-set by the parameters theta
% using a specified transformation type
%
% Campbell, D., and Petersson, L., “An Adaptive Data Representations for
% Robust Point-Set Registration and Merging”, International Conference on
% Computer Vision (ICCV), Santiago, Chile, IEEE, Dec. 2015
%
% For the full license, see license.txt in the root directory
%
% Author: Dylan Campbell
% Date: 20150709
% Revision: 5
%
% Input:
%   pointset: n x d static point-set (d: dimension)
%   transformType: {'rigid2d', 'rigid3d', 'affine2d', 'affine3d'}
%   theta: transformation parameters
% Output:
%   pointsetTransformed: n x d static point-set (d: dimension)

function [pointsetTransformed] = transformPointset(pointset, transformType, theta)

    switch lower(transformType)
        case 'rigid2d'
            pointsetTransformed = transformRigid2d(pointset, theta);
        case 'rigid3d'
            pointsetTransformed = transformRigid3d(pointset, theta);
        case 'affine2d'
            pointsetTransformed = transformAffine2d(pointset, theta);
        case 'affine3d'
            pointsetTransformed = transformAffine3d(pointset, theta);
        otherwise
            error('Unknown motion type');
    end
end

function [pointsetTransformed] = transformRigid2d(pointset, theta)

    n = size(pointset, 1);
    tx = theta(1);	% translation along the x dimension
    ty = theta(2);	% translation along the y dimension
    r = theta(3);	% rotation in radians
    
    R = [cos(r) -sin(r);
         sin(r)  cos(r)];
    
    pointsetTransformed(:, 1:2) = pointset(:, 1:2) * R' + ones(n, 1) * [tx ty];
end

function [pointsetTransformed] = transformRigid3d(pointset, theta)

    n = size(pointset, 1);
    tx = theta(5);	% translation along the x dimension
    ty = theta(6);	% translation along the y dimension
    tz = theta(7);	% translation along the z dimension
    
    R = quaternionToRotation(theta(1:4));
    
    pointsetTransformed(:, 1:3) = pointset(:, 1:3) * R' + ones(n, 1) * [tx ty tz];
end

function [pointsetTransformed] = transformAffine2d(pointset, theta)

    n = size(pointset,1);
    A = reshape(theta, 2, 3); % theta: [tx ty a11 a21 a12 a22]
    pointsetTransformed(:, 1:2) =  [ones(n, 1)  pointset(:, 1:2)] * A';
end

function [pointsetTransformed] = transformAffine3d(pointset, theta)

    n = size(pointset, 1);
    A = reshape(theta, 3, 4); % theta: [tx ty tz a11 a21 a31 a12 a22 a32 a13 a23 a33]
    pointsetTransformed(:, 1:3) =  [ones(n, 1)  pointset(:, 1:3)] * A';
end