%% Support Vector Registration (SVR)
% An algorithm for robust point-set registration using the Support Vector-
% parametrised Gaussian Mixture (SVGM) data representation
%
% Campbell, D., and Petersson, L., “An Adaptive Data Representations for
% Robust Point-Set Registration and Merging”, International Conference on
% Computer Vision (ICCV), Santiago, Chile, IEEE, Dec. 2015
%
% For the full license, see license.txt in the root directory
%
% Author: Dylan Campbell
% Date: 20150709
% Revision: 5
%
% Input:
%   scene: n x d static point-set (d: dimension)
%   model: m x d moving point-set (d: dimension)
%   cfg: (optional) handle to config function (e.g. @config2d)
% Output:
%   optTheta: optimal transformation parameters
%   durations: time elapsed [train scene SVM, train model SVM, align GMMs]

function [optTheta, durations] = svr(scene, model, cfg)

    %% Parse Config File
    % Set default config function
    if nargin < 3
        cfg = @config;
    end
    [p] = cfg();

    %% Load Point-Sets

    S = load(scene, '-ascii');
    M = load(model, '-ascii');

    d = size(S, 2); % Point-Set Dimension

    if (d ~= 2) && (d ~= 3)
        error('The point-sets must be 2D (n x 2) or 3D (n x 3).');
    end
    
    %% Estimate Scale

    if p.estimateSigma
        S_ = S - repmat(mean(S, 1), size(S, 1), 1); % Demean point-set
        [n, d] = size(S_);
        p.sigma = power(det(S_' * S_ / (n - 1)), 1 / (2 * d));
        p.gamma = 1 / (2 * power(p.sigma, 2));
    end

    %% Train One-Class SVMs

    % Set up training data
    nSVR = p.nSVR; % Maximum number of training points for SVM
    
    nS = size(S, 1);
    if nSVR < nS
        nSTrain = nSVR;
        STrainIndices = randperm(nS, nSTrain); % Random sampling
    else
        nSTrain = nS;
        STrainIndices = 1:nSTrain;
    end
    STrainData = S(STrainIndices, :);
    STrainLabel = ones(nSTrain, 1);

    nM = size(M, 1);
    if nSVR < nM
        nMTrain = nSVR;
        MTrainIndices = randperm(nM, nMTrain); % Random sampling
    else
        nMTrain = nM;
        MTrainIndices = 1:nMTrain;
    end
    MTrainData = M(MTrainIndices, :);
    MTrainLabel = ones(nMTrain, 1);

    % Train SVMs
    trainOption = sprintf('-s 2 -t 2 -g %f -n %f -m 1000 -e 0.001 -h 0 -b 0', p.gamma, p.nu);

    durations = zeros(1, 3);
    tic;
    SSVM = svmtrain(STrainLabel, STrainData, trainOption);
    durations(1) = toc;

    tic;
    MSVM = svmtrain(MTrainLabel, MTrainData, trainOption);
    durations(2) = toc;

    %% Map SVMs to GMMs

    muX = full(MSVM.SVs);
    phiX = MSVM.sv_coef * (2 * pi * p.sigma^2)^(d / 2);

    muY = full(SSVM.SVs);
    phiY = SSVM.sv_coef * (2 * pi * p.sigma^2)^(d / 2);

    p.muX = muX;
    p.phiX = phiX;
    p.muY = muY;
    p.phiY = phiY;

    %% L2 Registration

    % Set optimisation options
    [optimOptions] = setOptimOptions(p);

    x0 = p.initParam;
    
    % Convert sigma from SVGM sigma to SVR sigma
    % A smaller sigma can often produce a more refined registration result
    % but can be less stable
    p.sigma = p.sigma / p.sigmaFactor;

    tic;
    switch lower(p.constraint)
        case 'reproject'
            optTheta = fmincon(@svrObjFunc, x0, [], [], [], [], p.lb, p.ub, [], optimOptions, p);
        case 'unitnorm'
            optTheta = fmincon(@svrObjFunc, x0, [], [], [], [], p.lb, p.ub, @unit_norm, optimOptions, p);
        otherwise
            error('No rotation constraint specified');
    end
    durations(3) = toc;

    % To transform model point-set by theta:
    % Mt = transformPointset(M, p.transformType, optTheta);
end

function [optimOptions] = setOptimOptions(p)

    optimOptions = optimset('Algorithm', 'active-set', 'LargeScale','off');
    
    % Alternative optimisation algorithms:
    % {'trust-region-reflective', 'LargeScale','on'}
    % {'interior-point', 'LargeScale','on'}
    % {'sqp', 'LargeScale','off'}
    
    optimOptions = optimset(optimOptions, 'TolFun',1e-10, 'TolX',1e-10, 'TolCon', 1e-10);
    optimOptions = optimset(optimOptions, 'MaxFunEvals', p.maxIter, 'MaxIter', p.maxIter);
    optimOptions = optimset(optimOptions, 'GradObj', 'on');
    optimOptions = optimset(optimOptions, 'GradConstr','on');
    optimOptions = optimset(optimOptions, 'display','iter');
    
end

% Unit-norm constraint on quaternions
function [c, ceq, gc, gceq] = unit_norm(x, p)
    
    switch lower(p.transformType)
        case 'rigid3d'
            c = [];
            ceq = x(1)^2 + x(2)^2 + x(3)^2 + x(4)^2 - 1;
            if nargout > 2
                gc = [];
                gceq = 2 * [x(1); x(2); x(3); x(4); 0; 0; 0];
            end
        otherwise
            c = [];
            ceq = [];
            if nargout > 2
                gc = [];
                gceq = [];
            end
    end
end